#!/usr/bin/perl -w

use strict;
use Getopt::Std;
use IO::Socket; 
use Fcntl qw(:flock);

# Command line options:
# v=verbose, d=debug,h=help, r=remote host, c=check, i=init, s=sanity check
# n=new tarball, G=generate keys, V=version, N=no mail, C=ICU Config, H=Host config
use vars qw ( 
	       $opt_v $opt_d $opt_h $opt_r $opt_c $opt_i 
	       $opt_s $opt_n $opt_G $opt_V $opt_N $opt_C $opt_H 
            );

my $version = 'ICU v0.3 By Andreas stling, andreaso@it.su.se.';

my $icu_config = './ICU.conf';  	# ICU configuration file to use.	

my $verbose = 0;    			# 0/1 Always use verbose mode? (Used for testing)
my $debug   = 0;    			# 0/1 Always use debug mode?   (Used for testing)

# All sub routines we need
sub parse_command_line;
sub usage;
sub check_root;
sub read_icu_config;
sub check_port;
sub get_host_entry;
sub init_db;
sub create_new_tarball;
sub check_filesystem;
sub getfile;
sub getfiles;
sub lockfile;
sub send_mail;
sub get_mail;
sub asciifilecmp;
sub tmp_dir;
sub clean_exit;
sub notify;
sub beepbeepbeepbeep;
sub icu_die;                 
sub sanity_check;
sub generate_keys;
sub generate_key;
sub get_timestamp;

# Hashes we need.
my (%conf, %icuhost);

# Arrays we need.
my (@aide_output,@args);

# Scalars we need.
my ( 
     $mail_admin, $aide_conf_md5_differs, $scp_output, $i, 
     $errormsg, $hostname, $start_date, $subject, $mailbody, $timeout_reason,
     $old_date, $old_time, $new_date, $new_time, $tmp_dir
   ); 

my $init_done         = 0;		# The init is not done yet.
my $check_done        = 0;              # The check is not done yet.
my $invalid_timestamp = 0;		# The database creation timestamp is not invalid yet.

my $timeout_checking  = 0;		# One of these will be set to 1 after each alarm(n)
my $timeout_connect   = 0;		# so sub beepbeepbeepbeep knows what timed out.
my $timeout_getfiles  = 0;              # It will be set back 0 after each alarm(0).

$conf{'logger_arguments'} = "";		# Should be defined but empty. May be overwritten by
					# "logger_arguments = -foo" in ICU.conf.


#### MAIN ####

$SIG{'ALRM'} = 'beepbeepbeepbeep';	# Call beepbeepbeepbeep when something times out.

select(STDERR);				# No buffering.
$| = 1;
select(STDOUT);
$| = 1;

@args = @ARGV; 				# Save away the original arguments.
$start_date = (scalar localtime); 	# The time ICU was started.

parse_command_line;			# Parse the command line.
read_icu_config;			# Read ICU configuration file ($icu_config).

if (defined($opt_H)) { 		        # -H <file> => Use this file for host configuration.
    $conf{"icuhosts"} = $opt_H;
    $verbose && print "Using $conf{'icuhosts'} as host configuration file.\n";
}

generate_keys if (defined($opt_G));     # Generate keys and exit.
sanity_check;			        # Do a sanity check. If it succeeds it means that our
				        # client host info is now successfully stored in %icuhost.
	
$conf{"lockfile_dir"} =  $conf{'tmp_dir'};
$tmp_dir              = "$conf{'tmp_dir'}/$icuhost{'name'}";  # Give the host its own temporary directory.


if (defined($opt_n)) {                  # User specified -n, so create a new tarball.
    tmp_dir('new');
    create_new_tarball;
    clean_exit;
}

if (defined($opt_i)) {		        # User specified -i, so initialize a new database.
    lockfile("create"); 
    tmp_dir("new");
    notify(2,0,1,1,"Database initialization started.");

    if (check_port($icuhost{'name'},$icuhost{'sshport'})) {  
        init_db;
        if (defined($mailbody)) {
	    $debug && print "The mail body is not empty after init_db: not setting init_done to 1.\n";
	} 
	else {
	    $init_done = 1;		# Mark that the init is done so the welcome mail will be sent later.
	}
    } 
    else {
        notify(1,1,1,1,"Could not connect to port $icuhost{'sshport'}, aborting."); 
        get_mail("cant_connect");
        $subject  .= ": $errormsg";
        $mailbody .= "Additional information about the connection failure: $errormsg\n"; 	
    }                
    clean_exit;
}

if (defined($opt_c)) {              	# User specified -c so do a filesystem check.
    notify(0,0,1,1,"Check started.");
    lockfile("create");
    tmp_dir("new"); 

    if (check_port($icuhost{'name'},$icuhost{'sshport'})) {
        check_filesystem;
    } 
    else {
        notify(1,1,1,1,"Could not connect to port $icuhost{'sshport'}: $errormsg.");
        get_mail("cant_connect");
        $subject  .= ": $errormsg";
        $mailbody .= "Additional information about the connection failure: $errormsg\n";
        $mail_admin = 1 if ($conf{'mail_to_icu_admin_port_closed'});
    }

    clean_exit;
}


#### END OF MAIN ####



sub parse_command_line	
{
    $debug && print STDERR "* sub parse_command_line started\n";
    my $cmdline_ok = getopts('dvhinsr:cVNC:H:G');

    usage             if (defined($opt_h)); 				     # -h => show usage and exit.
    die("$version\n") if (defined($opt_V)); 				     # -V => show version and exit.

    usage unless (defined($opt_s) xor defined($opt_n) xor defined($opt_i) 
              xor defined($opt_c) xor defined($opt_G));   		     # Only one of these may be specified.

    $verbose = 1 if (defined($opt_v) || defined($opt_d) || defined($opt_i)); # Verbose if -v,-d or -i.
    $debug   = 1 if (defined($opt_d));    				     # -d => debug mode.
    $opt_N   = 1 if (defined($opt_d) || defined($opt_n));                    # Don't sent mail in debug mode. (no point
									     # since all output is printed to stdout/err).
									     # And not when running -n neither.
  # -r must be specified unless running with -s or -G.
    if (!defined($opt_r) && !defined($opt_s) && !defined($opt_G)) {
	print STDERR "You must specify the hostname with -r <hostname>\n"; 
	$cmdline_ok = 0;
    } 

    $hostname = $opt_r if (defined($opt_r));				     # Grab the hostname.

    print "Verbose mode activated.\n" if ($verbose);
    print STDERR "Debug mode activated.\n" if ($debug);

    if (defined($opt_C)) {                   			 	     # Use another ICU configuration file.
	$icu_config = $opt_C;
	$verbose && print "Using $icu_config as ICU configuration file.\n";
    }               

    usage unless ($cmdline_ok);  					     # Something was wrong.

    $debug && print STDERR "* sub parse_command_line ended\n"; 
}                    



sub usage		# Print usage help and exit.
{
    select(STDERR);  
    $debug && print "* sub usage started\n"; 
    print "$version\n";
    print "General usage:\n$0 -r <hostname> <action> [options]\n";
    print "Or sanity check:\n$0 -s [-r <hostname>] [options]\n";
    print "Or key generation:\n$0 -G\n\n";
    print "Actions:\n";
    print "-n                     Create tarball for a new client.\n";
    print "-i                     Initialize new database.\n";
    print "-c                     Do a filesystem check.\n";
    print "Options:\n";
    print "-C <ICU config>        Use <ICU config> as ICU configuration file.\n";
    print "-H <host config>       Use <host config> as host configuration file.\n";
    print "-N                     Print mail to standard output instead of sending it.\n"; 
    print "-v                     Verbose mode.\n";
    print "-d                     Debug mode - lots of output. AIDE output goes to stdout\n";
    print "                       and scp/ssh is run in verbose mode. No mail will be sent.\n";
    print "Other:\n";
    print "-r <hostname>          Host to work with. Must be one of the hosts in ICU.hosts.\n";
    print "-s                     Just do a sanity check and exit.\n";
    print "-G                     Generate keys.\n"; 
    print "-h                     Show this help.\n";
    print "-V                     Just show the version and quit.\n"; 
    select(STDOUT);
    exit;
}



sub read_icu_config		# Read $icu_config and put parameters and their values in %conf with 
				# the parameter name as key.
{
    $debug && print STDERR "* sub read_icu_config started\n";
    $debug && print STDERR "Reading ->$icu_config<-\n";      

    my (@cfg, @tmparr, $param, $value, $line, $key);

    unless (-e "$icu_config") {
	print STDERR "Fatal error: Could not find $icu_config.\n";
	print STDERR "You must start ICU from inside the ICU directory or specify which\n";
	print STDERR "configuration file to use with the -C <config> argument.\n";
	exit;
    }

    open(CFG,$icu_config) or die("CFG: Could not open $icu_config: $!\n\nExiting");
    @cfg = <CFG>;
    close(CFG);

    for ($i = 0; $i <= $#cfg; $i++) {
	chomp($cfg[$i]);
	$cfg[$i] =~ s/^\s+//;          			 # Remove leading whitespaces.
	$cfg[$i] =~ s/\s*\#.*//;       			 # Remove #comments.
	$cfg[$i] =~ s/\s+$//; 	       			 # Remove trailing whitespaces.
	$cfg[$i] =~ s/\s*=\s*/=/;      			 # Remove whitespaces next to the first =.

	($param, $value) = ($cfg[$i] =~ /(.+?)=(.+)/);	 # Try to grab parameter and its value.

	if (defined($param) && defined($value)) {
	    $param = lc ($param);
	    $conf{"$param"} = $value;			 # Save in %conf.
	    $debug && print STDERR "Parameter = ->$param<-, value = ->$conf{$param}<-\n"; 
	}                       
	else {  					 # Not a valid x=y line.
	    $line = $i + 1;
	    print STDERR "Warning: $icu_config line $line is invalid - ignoring.\n" 
	      if ($cfg[$i] =~ /\S+/ && ($verbose || defined($opt_s)));
	}
    } 

  # Replace %icu_basedir% with $conf{'icu_basedir'} and %host_basedir% with 
  # $conf{'host_basedir'} in all variables.
    foreach $key (keys(%conf)) {
	$conf{"$key"} =~ s/%icu_basedir%/$conf{'icu_basedir'}/g;
	$conf{"$key"} =~ s/%host_basedir%/$conf{'host_basedir'}/g;
    }

    $debug && print STDERR "* sub read_icu_config ended\n"; 
} 



sub get_host_entry {  # Get host from $conf{'icuhosts'}. 
		      # Syntax: get_host_entry("hostname"). Values will be stored in %icuhost.

    $debug && print STDERR "* sub get_host_entry started\n"; 

    my (@icuhosts, $line, $host);
    $host = shift;
    $debug && print STDERR "Searching for $host in $conf{'icuhosts'}.\n";

    open(ICUHOSTS,$conf{'icuhosts'}) or die("ICUHOSTS: Could not open $conf{'icuhosts'}: $!\n\nExiting");
    @icuhosts = <ICUHOSTS>;
    close(ICUHOSTS);

    for ($i = 0; $i <= $#icuhosts; $i++) {
	$icuhosts[$i] =~ s/\s*\#.*//;       			# Remove #comments.
	$icuhosts[$i] =~ s/\s+//g;				# Remove all whitespaces.

	if ($icuhosts[$i] =~ /^\S+:\S+:\S+:\S+:\d{1,5}$/) {     # Correct host entry. 
  	    ($icuhost{"name"},$icuhost{"mailto"},$icuhost{"os"},$icuhost{"arch"}, $icuhost{"sshport"}) 
	      = split(/:/,$icuhosts[$i],5);
		
	    if ($icuhost{'name'} eq $host) {    		# It is the host we are looking for.
		$debug && print STDERR "* sub get_host_entry returning 1\n";
		return 1;
	    }
	} 
	else {							# Incorrect host entry.
            $line = $i + 1;
	    print STDERR "Warning: $conf{'icuhosts'} line $line is invalid - ignoring.\n" 
	      if ($icuhosts[$i] =~ /\S+/ && ($verbose || defined($opt_s)));
	}           
    } 

    undef(%icuhost);
    $debug && print STDERR "* sub get_host_entry returning 0\n";
    return 0;
}



sub init_db    # Initialize a new database.
{
    $debug && print STDERR "* sub init_db started\n";

    my ($tmpstr, $date, $error, $fromfile, $tofile, $dbdate, $dbtime); 

  # Run SSH in verbose mode if ICU is running in debug mode.
    if ($debug) {
	print STDERR "Adding -v flag to SSH command.\n";
	$conf{"ssh"} =~ s/$conf{'ssh'}/$conf{'ssh'} -v/; 
    }

  # Replace %stuff% with $conf{'stuff'}. 
    $conf{"init_db_cmd"} =~ s/%key_init_db_priv%/$conf{'key_init_db_priv'}/g;
    $conf{"init_db_cmd"} =~ s/%port%/$icuhost{'sshport'}/g;	
    $conf{"init_db_cmd"} =~ s/%hostname%/$icuhost{'name'}/g;
    $conf{"init_db_cmd"} =~ s/%ssh%/$conf{'ssh'}/g;
    $conf{"init_db_cmd"} =~ s/%aide_init_db_cmd%/\"$conf{'aide_init_db_cmd'}\"/g;

  # Now execute the AIDE db init command on the remote host.
  # The right command will be automatically executed on the remote host 
  # because of the SSH key (unless the root acount doesn't have a password).
  # We pass the command to SSH just in case.
    print "Executing init command: '$conf{'init_db_cmd'}'\n";
    print "This may take a while.\n";
    system("$conf{'init_db_cmd'}");

    $date = (scalar localtime);
    getfiles("db","conf");		# Get first database and configuration.

  # Was the database successfully created? Let's check if there's a creation timestamp in it.
    print "Looking for database creation timestamp...";
    ($dbdate,$dbtime) = get_timestamp("$tmp_dir/aide.db");
    if (defined($dbdate) && defined($dbtime)) {
	print "found - database successfully initialized.\n";
	notify(0,0,1,1,"The database was successfully initialized.\n");
    }
    else {
	print "NOT found.\n";		# Database creation timestamp was NOT found.
	notify(2,0,0,1,"$mailbody") if (defined($mailbody));	
	notify(2,0,1,1,"Could not find a creation timestamp in the database $tmp_dir/aide.db.\n");
	print "Something must have gone really wrong on the remote host.\n";
	print "Please check out why $tmp_dir/aide.db doesn't contain a creation timestamp.\n";  
	notify(2,0,1,1,"The database was NOT initialized.\nExiting...");
	lockfile("remove");   	        # Remove lock file but keep the tmp files so you can take a look at them.
	exit;
    }

  # Move aide.db to the host's directory and rename it to aide.db.current.
    $fromfile = "$tmp_dir/aide.db";
    $tofile   = "$conf{'database_dir'}/$icuhost{'name'}/aide.db.current";
    notify(2,0,0,1,"Saving database as $tofile.\n");
    rename("$fromfile","$tofile") || icu_die("Rename $fromfile to $tofile: $!");

  # Also make a copy of aide.db.current to archive/aide.db-first-$date.
    $fromfile = "$conf{'database_dir'}/$icuhost{'name'}/aide.db.current";
    $tofile   = "$conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-first-$date";
    notify(2,0,0,1,"Saving copy as $tofile.\n");
    ($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && icu_die("Copy $fromfile to $tofile: $error"); 

  # Compress the database.
    if ($conf{'compress_databases'}) {
	print "Compressing database...";
        system("$conf{'compress'} \"$conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-first-$date\" >/dev/null");
	print "done.\n";
    }

  # Move aide.conf to the host's directory and rename it to aide.conf.current.
    $fromfile = "$tmp_dir/aide.conf";
    $tofile   = "$conf{'database_dir'}/$icuhost{'name'}/aide.conf.current";
    notify(2,0,0,1,"Saving configuration as $tofile.\n");
    rename("$fromfile","$tofile") || icu_die("Rename $fromfile to $tofile: $!");

  # Also make a copy of aide.conf.current to archive/aide.conf-first-$date.
    $fromfile = "$conf{'database_dir'}/$icuhost{'name'}/aide.conf.current";
    $tofile   = "$conf{'database_dir'}/$icuhost{'name'}/archive/aide.conf-first-$date";
    notify(2,0,0,1,"Saving copy as $tofile.\n");
    ($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && icu_die("Copy $fromfile to $tofile: $error");

    $debug && print STDERR "* sub init_db ended\n";
}



sub create_new_tarball		# Create client tarball.
{
    $debug && print STDERR "* sub create_new_tarball started\n"; 

    my ( 
         @instscript, $tmpstr, $key_check_pub_path, $key_init_db_pub_path, $key_get_bin_pub_path,
         $key_get_conf_pub_path, $key_get_db_pub_path, $fromfile, $tofile, $error, $files
       );

    notify(2,0,1,1,"Creating new tarball for $icuhost{'name'}. OS=$icuhost{'os'}, architecture=$icuhost{'arch'}.");

  # Create temporary directory where we keep the files that should be in the tarball.
    $tmpstr = "$tmp_dir/$icuhost{'name'}.icu-install";
    mkdir("$tmpstr",0700) or die("Fatal error: Could not create $tmpstr: $!\n\nExiting"); 

  # Create directories where the client's databases/configurations etc will be kept.
    $tmpstr = "$conf{'database_dir'}/$icuhost{'name'}";
    mkdir("$tmpstr",0700) or die("Fatal error: Could not create $tmpstr: $!\n\nExiting");
    $tmpstr .= "/archive";
    mkdir("$tmpstr",0700) or die("Fatal error: Could not create $tmpstr: $!\n\nExiting"); 

  # Copy AIDE binary to the temporary directory.
    $fromfile = "$conf{'bin_dir'}/aide.bin-$icuhost{'arch'}-$icuhost{'os'}";
    $tofile   = "$tmp_dir/$icuhost{'name'}.icu-install/aide.bin";
    ($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && die("Could not copy $fromfile to $tofile: $error\nExiting");

  # Copy AIDE configuration to the temporary directory.
    $fromfile = "$conf{'config_dir'}/aide.conf-$icuhost{'arch'}-$icuhost{'os'}";
    $tofile   = "$tmp_dir/$icuhost{'name'}.icu-install/aide.conf";
    ($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && die("Could not copy $fromfile to $tofile: $error\nExiting");

  # Copy the init_db key.
    $key_init_db_pub_path = $conf{'key_init_db_pub'};    # Path and filename here.
    $conf{'key_init_db_pub'} =~ s/.+\///;                # Only the filename here.
    $fromfile = "$key_init_db_pub_path";
    $tofile   = "$tmp_dir/$icuhost{'name'}.icu-install/$conf{'key_init_db_pub'}";
    ($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && die("Could not copy $fromfile to $tofile: $error\nExiting");

  # Copy the check key.
    $key_check_pub_path = $conf{'key_check_pub'};
    $conf{'key_check_pub'} =~ s/.+\///;
    $fromfile = "$key_check_pub_path";
    $tofile   = "$tmp_dir/$icuhost{'name'}.icu-install/$conf{'key_check_pub'}"; 
    ($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && die("Could not copy $fromfile to $tofile: $error\nExiting");

  # Copy the get_bin key.
    $key_get_bin_pub_path = $conf{'key_get_bin_pub'};
    $conf{'key_get_bin_pub'} =~ s/.+\///;
    $fromfile = "$key_get_bin_pub_path";
    $tofile   = "$tmp_dir/$icuhost{'name'}.icu-install/$conf{'key_get_bin_pub'}";
    ($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && die("Could not copy $fromfile to $tofile: $error\nExiting");

  # Copy the get_conf key.
    $key_get_conf_pub_path = $conf{'key_get_conf_pub'};
    $conf{'key_get_conf_pub'} =~ s/.+\///;
    $fromfile = "$key_get_conf_pub_path";
    $tofile   = "$tmp_dir/$icuhost{'name'}.icu-install/$conf{'key_get_conf_pub'}";
    ($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && die("Could not copy $fromfile to $tofile: $error\nExiting");

  # Copy the get_db key.
    $key_get_db_pub_path = $conf{'key_get_db_pub'};
    $conf{'key_get_db_pub'} =~ s/.+\///;
    $fromfile = "$key_get_db_pub_path";
    $tofile   = "$tmp_dir/$icuhost{'name'}.icu-install/$conf{'key_get_db_pub'}";
    ($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && die("Could not copy $fromfile to $tofile: $error\nExiting");

  # Copy the install script README.
    $fromfile = "$conf{'readme_installscript'}";
    $tofile   = "$tmp_dir/$icuhost{'name'}.icu-install/README";
    ($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && die("Could not copy $fromfile to $tofile: $error\nExiting");

  # Open the install script and replace %stuff% with $conf{'stuff'}.
    open(INSTSCRIPT,$conf{'icu_client_installscript'}) 
      or die "INSTSCRIPT: Could not open $conf{'icu_client_installscript'}: $!\n\nExiting";
    @instscript = <INSTSCRIPT>;
    close(INSTSCRIPT);

  # Replace %stuff% with $conf{'stuff'} in the client install script.
    for ($i = 0; $i <= $#instscript; $i++) {
	$instscript[$i] =~ s/%host_basedir%/$conf{'host_basedir'}/g;		
	$instscript[$i] =~ s/%key_check_pub%/$conf{'key_check_pub'}/g; 
	$instscript[$i] =~ s/%key_init_db_pub%/$conf{'key_init_db_pub'}/g;
	$instscript[$i] =~ s/%key_get_bin_pub%/$conf{'key_get_bin_pub'}/g;  
	$instscript[$i] =~ s/%key_get_conf_pub%/$conf{'key_get_conf_pub'}/g; 
	$instscript[$i] =~ s/%key_get_db_pub%/$conf{'key_get_db_pub'}/g; 
	$instscript[$i] =~ s/%icu_server_name%/$conf{'icu_server_name'}/g;
	$instscript[$i] =~ s/%admin_email%/$conf{'admin_email'}/g;
	$instscript[$i] =~ s/%hostname%/$icuhost{'name'}/g;
	$instscript[$i] =~ s/%arch%/$icuhost{'arch'}/g;  
	$instscript[$i] =~ s/%os%/$icuhost{'os'}/g;
    }

  # Write the install script to disk.
    $tmpstr = "$tmp_dir/$icuhost{'name'}.icu-install/icu-install.sh";
    open(INSTSCRIPT,">$tmpstr") or die "INSTSCRIPT: Could not open $tmpstr for writing: $!\n\nExiting";
    print INSTSCRIPT @instscript;
    close(INSTSCRIPT);
    chmod(0700,"$tmp_dir/$icuhost{'name'}.icu-install/icu-install.sh");

  # Now create the tarball.
    chdir("$tmp_dir") 
      or die("Fatal error, could not change to working directory $tmp_dir: $!\n\nExiting");
    $tmpstr = "$conf{'database_dir'}/$icuhost{'name'}/$icuhost{'name'}.icu-install.tar";
    system("$conf{'tar'}","cf","$tmpstr","$icuhost{'name'}.icu-install");

  # Make sure the tarball was actually created.
    if (-e "$tmpstr") {	 
	notify(2,0,1,1,"Saved tarball as $tmpstr.");
	print "This file should now be untared and installed on the host $icuhost{'name'}.\n";
    }
    else {
	print STDERR "Something went wrong - could not create tarball.\n";
	print STDERR "Inspect previous error messages and try again.\n";
    }

    $debug && print STDERR "* sub create_new_tarball ended\n"; 
}



sub check_filesystem	# Do a filesystem check.
{
    $debug && print STDERR "* sub check_filesystem started\n";   

    my ( 
	 @files, @missing_files, $tmpstr, $date, $local_md5sum, $error,
         $remote_md5sum,$added, $removed, $fromfile, $tofile
       );

  # Get binary, configuration and database from the remote host so we can verify the MD5 fingerprints.
    getfiles("bin","conf","db");    # Get aide, aide.conf and aide.db.

  # Verify MD5 fingerprint of aide.db.
    $verbose && print "Verifying MD5 fingerprint of the AIDE database...";

    ($local_md5sum)  = (`$conf{'md5sum'} $conf{'database_dir'}/$icuhost{'name'}/aide.db.current` =~ /([\da-f]{32})/);
    ($remote_md5sum) = (`$conf{'md5sum'} $tmp_dir/aide.db`                                       =~ /([\da-f]{32})/);

    icu_die("Failed to calculate MD5 fingerprint of aide.db.current/aide.db!!")
      unless (defined($local_md5sum) && defined($remote_md5sum));

    if ($local_md5sum eq $remote_md5sum) {              # MD5 ok.
	$verbose && print "match.\n";
    } 
    else {						# MD5 NOT ok.
	$date = (scalar localtime);
	$verbose && print "MISMATCH.\n"; 
	notify(1,0,1,1,"The MD5 fingerprint of aide.db did NOT match!");
	get_mail("aide_db_md5_mismatch");
	$mail_admin = 1 if ($conf{'mail_to_icu_admin_db_md5'});  

  # Incorrect MD5 fingerprint of aide.db - Check what had added/removed.
        $verbose && print "Finding differences between databases, be patient...";
        ($added,$removed) =
          asciifilecmp("$conf{'database_dir'}/$icuhost{'name'}/aide.db.current","$tmp_dir/aide.db");
        $verbose && print "done.\n";

        $mailbody .= "Additional information:\n";
        $mailbody .= "Fingerprint of locally stored database: $local_md5sum\n";
        $mailbody .= "Fingerprint of remote database:         $remote_md5sum\n";
 
  # Print size of locally stored database.
        ($tmpstr) = (stat("$conf{'database_dir'}/$icuhost{'name'}/aide.db.current"))[7]
	  || icu_die ("Unable to stat $conf{'database_dir'}/$icuhost{'name'}/aide.db.current");
        $mailbody .= "Size of locally stored database: $tmpstr\n";
 
  # Print size of remote database:
        ($tmpstr) = (stat("$tmp_dir/aide.db"))[7]
	  || icu_die ("Unable to stat $tmp_dir/aide.db");
        $mailbody .= "Size of remote database:         $tmpstr\n";

  # Update database/hostname/aide.db.current.
        notify(1,0,1,1,"Updating $conf{'database_dir'}/$icuhost{'name'}/aide.db.current.");
	$fromfile = "$tmp_dir/aide.db";
	$tofile   = "$conf{'database_dir'}/$icuhost{'name'}/aide.db.current";
  # Copy the database, but don't remove it because we'll use 'get_timestamp' on it below.
	($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && icu_die("Copy $fromfile to $tofile: $error");

  # Save copy in archive/.
        notify(1,0,1,1,"Saving copy as $conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-$date.");
	$fromfile = "$conf{'database_dir'}/$icuhost{'name'}/aide.db.current";
	$tofile   = "$conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-$date";
        ($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && icu_die("Copy $fromfile to $tofile: $error");

        system("$conf{'compress'}","$conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-$date")
          if ($conf{'compress_databases'});

	if (defined($added)) {			# Something had been added to aide.db.
	    $mailbody .= "\nThe following line(s) had been added to $conf{'host_basedir'}/aide.db ";
	    $mailbody .= "on $icuhost{'name'}:\n$added\n";
	}

	if (defined($removed)) { 		# Something had been removed from aide.db.
	    $mailbody .= "\nThe following line(s) had been removed from $conf{'host_basedir'}/aide.db "; 
	    $mailbody .= "on $icuhost{'name'}:\n$removed\n";  
	}
    }

  # Verify the MD5 fingerprint of aide.conf.
    $verbose && print "Verifying MD5 fingerprint of the AIDE configuration..."; 

    ($local_md5sum)  = (`$conf{'md5sum'} $conf{'database_dir'}/$icuhost{'name'}/aide.conf.current` =~ /([\da-f]{32})/);
    ($remote_md5sum) = (`$conf{'md5sum'} $tmp_dir/aide.conf`                                       =~ /([\da-f]{32})/);  

    icu_die("Failed to calculate MD5 fingerprint of aide.conf.current/aide.conf!!")
      unless (defined($local_md5sum) && defined($remote_md5sum));
	
    if ($local_md5sum eq $remote_md5sum) {   
	$verbose && print "match.\n";           # MD5 ok.
    } 
    else {					# MD5 NOT ok.
	$verbose && print "MISMATCH.\n";       
	$date = (scalar localtime); 
	$aide_conf_md5_differs = 1;
	notify(1,0,1,1,"The MD5 fingerprint of aide.conf did NOT match!");
	get_mail("aide_conf_md5_mismatch");
	$mail_admin = 1 if ($conf{'mail_to_icu_admin_conf_md5'});

  # Incorrect MD5 fingerprint of aide.conf - Check what had added/removed.
        $verbose && print "Finding differences between configurations...";
        ($added,$removed) =
          asciifilecmp("$conf{'database_dir'}/$icuhost{'name'}/aide.conf.current","$tmp_dir/aide.conf");
        $verbose && print "done.\n";

        $mailbody .= "Additional information:\n";
        $mailbody .= "Fingerprint of locally stored configuration:  $local_md5sum\n";
        $mailbody .= "Fingerprint of remote configuration:          $remote_md5sum\n";
 
  # Print size of locally stored config.
        ($tmpstr) = (stat("$conf{'database_dir'}/$icuhost{'name'}/aide.conf.current"))[7]
	  || icu_die ("Unable to stat database_dir'}/$icuhost{'name'}/aide.conf.current");
        $mailbody .= "Size of locally stored configuration: $tmpstr\n";
 
  # Print size of remote config.
        ($tmpstr) = (stat("$tmp_dir/aide.conf"))[7]
	  || icu_die ("Unable to stat $tmp_dir/aide.conf");
        $mailbody .= "Size of remote configuration:         $tmpstr\n";

  # Update database/hostname/aide.conf.current.
        notify(1,0,1,1,"Updating $conf{'database_dir'}/$icuhost{'name'}/aide.conf.current.");
	$fromfile = "$tmp_dir/aide.conf";
	$tofile   = "$conf{'database_dir'}/$icuhost{'name'}/aide.conf.current";
        rename("$fromfile","$tofile") || icu_die("Rename $fromfile to $tofile: $!");

  # Save copy in archive/.
        notify(1,0,1,1,"Saving copy as $conf{'database_dir'}/$icuhost{'name'}/archive/aide.conf-$date.");
	$fromfile = "$conf{'database_dir'}/$icuhost{'name'}/aide.conf.current";
	$tofile   = "$conf{'database_dir'}/$icuhost{'name'}/archive/aide.conf-$date";
        ($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && icu_die("Copy $fromfile to $tofile: $error");

	if (defined($added)) {                  # Something had been added to aide.conf.
	    $mailbody .= "\nThe following line(s) had been added to $conf{'host_basedir'}/aide.conf ";
	    $mailbody .= "on $icuhost{'name'}:\n$added\n";
	}

	if (defined($removed)) {                # Something had been removed from aide.conf.
	    $mailbody .= "\nThe following line(s) had been removed from $conf{'host_basedir'}/aide.conf ";
	    $mailbody .= "on $icuhost{'name'}:\n$removed\n";
	}
    }                

  # Verify the MD5 fingerprint of the AIDE binary.
    $verbose && print "Verifying MD5 fingerprint of the AIDE binary...";
 
    ($local_md5sum)  = (`$conf{'md5sum'} $conf{'bin_dir'}/aide.bin-$icuhost{'arch'}-$icuhost{'os'}` =~ /([\da-f]{32})/);
    ($remote_md5sum) = (`$conf{'md5sum'} $tmp_dir/aide.bin`                                         =~  /([\da-f]{32})/);

    icu_die("Failed to calculate MD5 fingerprint of aide.bin/aide.bin-$icuhost{'arch'}-$icuhost{'os'}!!")
      unless (defined($local_md5sum) && defined($remote_md5sum));

    if ($local_md5sum eq $remote_md5sum) {         # MD5 ok.
	$verbose && print "match.\n";
    } 
    else {                                         # MD5 NOT ok.
	$verbose && print "MISMATCH.\n";
	notify(1,0,1,1,"The MD5 fingerprint of the AIDE binary did NOT match!");
        $date = (scalar localtime);

	get_mail("aide_binary_md5_mismatch");
	$mail_admin = 1; # Always send mail to admin when the binary has changed

	$mailbody .= "Additional information:\n";
	$mailbody .= "Fingerprint of locally stored binary:  $local_md5sum\n";
	$mailbody .= "Fingerprint of remote binary:          $remote_md5sum\n";

  # Print size of locally stored binary.
        ($tmpstr) = (stat("$conf{'bin_dir'}/aide.bin-$icuhost{'arch'}-$icuhost{'os'}"))[7]
	  || icu_die ("Unable to stat $conf{'bin_dir'}/aide.bin-$icuhost{'arch'}-$icuhost{'os'}");
        $mailbody .= "Size of locally stored binary: $tmpstr\n";

  # Print size of remote binary.
	($tmpstr) = (stat("$tmp_dir/aide.bin"))[7]
	  || icu_die ("Unable to stat $tmp_dir/aide.bin");
	$mailbody .= "Size of remote binary:         $tmpstr\n";

        notify(1,0,1,1,"Saving remote mismatching AIDE binary to $conf{'database_dir'}/$icuhost{'name'}/archive/aide.bin-md5-mismatch-$date.");
	$fromfile = "$tmp_dir/aide.bin";
	$tofile   = "$conf{'database_dir'}/$icuhost{'name'}/archive/aide.bin-md5-mismatch-$date";
        ($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && icu_die("Copy $fromfile to $tofile: $error");

  # Should be continue even if the binary has changed?!
	if ($conf{'continue_if_binary_changed'}) {
	    $subject .= " - Check done anyway."; 
	    notify(1,0,1,1,"WARNING, continuing anyway.");
        } 
	else {
	    notify(1,0,1,1,"Check aborted.");
	    clean_exit;
	}
    }

    $debug && print STDERR "Setting timeout to $conf{'check_timeout'} seconds.\n";
    alarm($conf{'check_timeout'});
    $timeout_checking = 1;
	
  # Run SSH in verbose mode if ICU is running in debug mode.
    if ($debug) {
	print STDERR "Adding -v flag to SSH.\n";
	$conf{"ssh"} =~ s/$conf{'ssh'}/$conf{'ssh'} -v/;  
    }

  # Run SSH in quiet mode if ICU isn't running in debug or verbose mode.
    $conf{"ssh"} =~ s/$conf{ssh}/$conf{ssh} -q/ unless ($debug || $verbose);

  # Replace %stuff% with $conf{'stuff'} in the check commands.
    $conf{"check_cmd"} =~ s/%key_check_priv%/$conf{'key_check_priv'}/g;
    $conf{"check_cmd"} =~ s/%port%/$icuhost{'sshport'}/g;
    $conf{"check_cmd"} =~ s/%hostname%/$icuhost{'name'}/g; 
    $conf{"check_cmd"} =~ s/%ssh%/$conf{'ssh'}/;               
    $conf{"check_cmd"} =~ s/%aide_check_cmd%/\"$conf{'aide_check_cmd'}\"/;

  # Get old database creation timestamp.
   ($old_date,$old_time) = get_timestamp("$tmp_dir/aide.db");

  # Die if it wasn't there.
    icu_die("Could not get timestamp from old db $tmp_dir/aide.db")
      unless (defined($old_date) && defined($old_time));        

  # Now execute the real check.
    $debug && print STDERR "Executing $conf{'check_cmd'}\n"; 
    $verbose && print "Executing AIDE check command on the remote host. This may take a while.\n";
    @aide_output = "";

    if ($debug) {			  # Debug mode => give output in realtime.
	print STDERR "Executing AIDE check command using 'system'.\n";
	system("$conf{'check_cmd'}");         
    }
    else {			          # Not debug mode => save (possible) output for later use.
	open(CHECK,"$conf{'check_cmd'}|");
        while ($_ = <CHECK>) {		  # Filter out some not-so-very-interesting things.
	    chomp;
	    $debug && print STDERR "Got output. $_\n";
	    if (($_ =~ /^traverse_tree\(\):no such file or directory: .+$/i && $conf{'ignore_traverse_tree_warnings'})
	     || ($_ =~ /^warning: executing.*for ssh1 compatibility$/i)
             || ($_ =~ /^\/.+ atime in future$/ && $conf{'ignore_future_atimes'}) 
             || ($_ =~ /^\/.+ ctime in future$/ && $conf{'ignore_future_ctimes'})
	     || ($_ =~ /^\/.+ mtime in future$/ && $conf{'ignore_future_mtimes'})) {
		print "Suppressing warning: $_\n"		 	  # Print the warnings above if running a verbose check
		  if ($verbose && !$conf{'silent_suppressing'});          # and 'silent_suppressing' isn't 1.
	    }
	    else {
		push(@aide_output ,"$_\n");
	    }
        }
        close(CHECK);
    }

  # The check didn't time out so stop the alarm clock and restore the original mail.
    alarm(0);
    $timeout_checking = 0;
    $check_done       = 1;

    if ($#aide_output == 0) {                 # No output from AIDE.
	if ($conf{'allow_empty_reports'}) {   # Send report telling that no change was found?
	    get_mail("empty_report");
	} 
	else {
	    notify(1,0,1,1,"No output from AIDE -> No change in the filesystem was found.");
	}
    } 
    else {			              # A change in the filesystem was found.
	get_mail("filesystem_changed");  
	foreach $_ (@aide_output) {           # Show summary in the subject.
	    $subject .= " (Added=$1,"    if ($_ =~ /added files=(\d+)/i);
	    $subject .= "Removed=$1,"    if ($_ =~ /removed files=(\d+)/i); 
	    $subject .= "Changed=$1)"    if ($_ =~ /changed files=(\d+)/i); 
        }
	$subject  .= ". ";
	$mailbody .= "The check was started $start_date, and below is the output\n";
	$mailbody .= "from AIDE. Read it carefully.\n\n@aide_output\n\n";
    }                                              

  # Get the database and verify that the new timestamp is more recent than it was before the check.
    getfiles("db");

    $verbose && print "Verifying that the new database creation timestamp is more recent...";
    ($new_date,$new_time) = get_timestamp("$tmp_dir/aide.db"); 

  # Die if there was no timestamp in the database.
    icu_die("Could not get timestamp from new db $tmp_dir/aide.db")
      unless (defined($new_date) && defined($new_time));

    if ($new_time <= $old_time && $new_date <= $old_date) {
	$verbose && print "FAILED.\n(Timestamp after the check was NOT more recent than before)\n";

	$mail_admin = 1 if ($conf{'mail_icu_admin_invalid_timestamp'});
	get_mail("invalid_timestamp");
	$mailbody .= "Additional information:\n";
	$mailbody .= "Database creation timestamp before the check: $old_date $old_time.\n";
	$mailbody .= "Database creation timestamp after the check:  $new_date $new_time.\n";

	$debug && print STDERR "Setting invalid_timestamp to 1\n";
	$invalid_timestamp = 1;
    }
    else {
        $verbose && print "passed.\n";
    }

  # Update aide.db.current on the ICU server. 
    $fromfile = "$tmp_dir/aide.db"; 
    $tofile   = "$conf{'database_dir'}/$icuhost{'name'}/aide.db.current";
    rename("$fromfile","$tofile") || icu_die("Rename $fromfile to $tofile: $!");

  # Also save a copy of aide.db.current in the archive directory if a change in the filesystem was found.
    if ($#aide_output != 0 || $invalid_timestamp) { 
	$date = (scalar localtime);

	notify(1,0,1,1,"A change in the filesystem was found.") 
	    if ($#aide_output != 0);
	notify(1,0,1,1,"Invalid database creation timestamp! (old=$old_date $old_time, new=$new_date $new_time)")      
	    if ($invalid_timestamp);
	notify(1,0,1,1,"Updating $conf{'database_dir'}/$icuhost{'name'}/aide.db.current.");
	notify(1,0,1,1,"Saving copy as $conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-$date.");

	$fromfile = "$conf{'database_dir'}/$icuhost{'name'}/aide.db.current";
	$tofile   = "$conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-$date";
	($error = `$conf{'cp'} \"$fromfile\" \"$tofile\" 2>&1`) && icu_die("Copy $fromfile to $tofile: $error"); 

  # Compress the database.
	system("$conf{'compress'}","$conf{'database_dir'}/$icuhost{'name'}/archive/aide.db-$date")
	  if ($conf{'compress_databases'});
    }

    $debug && print STDERR "* sub check_filesystem ended\n";
}



sub getfile    # Get file from the remote host.  Syntax: getfile(command,filename);
	       # Clean_exit will be called if the file transfer fails.
{
    $debug && print STDERR "* sub getfile started\n"; 

    my $cmd      = shift;
    my $filename = shift;

    if ($verbose) {			   # Verbose mode => give (possible) output in realtime. 
        print " $filename";
        $debug && print STDERR "\nExecuting $cmd\n";
        system("$cmd");    
    }
    else {				   # Not verbose mode => save (possible) output for later use.
        open(GET,"$cmd|");
        while ($_ = <GET>) {
            $scp_output .= $_ unless ($_ =~ /executing scp1 compatibility/i);  # Comp. warnings are no fun.
        }
        close(GET);
    }

    $debug && print STDERR "* sub getfile ended\n"; 
}



sub getfiles   # Get file(s) from the remote host.  Syntax: getfiles("db"||"conf"||"bin").
{
    $debug && print STDERR "* sub getfiles started\n"; 

    my (@tmparr, @files, @missing_files);
    my $get_db   = 0;
    my $get_bin  = 0;
    my $get_conf = 0;

    $get_bin = 1 				# Should we get the binary?
        if (defined($_[0]) && $_[0] eq "bin" 
         || defined($_[1]) && $_[1] eq "bin" 
         || defined($_[2]) && $_[2] eq "bin");
  
    $get_conf = 1 				# Should we get the config?
	if (defined($_[0]) && $_[0] eq "conf" 
	 || defined($_[1]) && $_[1] eq "conf" 
	 || defined($_[2]) && $_[2] eq "conf");

    $get_db = 1					# Should we get the database?
	if (defined($_[0]) && $_[0] eq "db" 
	 || defined($_[1]) && $_[1] eq "db" 
	 || defined($_[2]) && $_[2] eq "db"); 

    if ($debug) {				# Run scp in verbose mode if ICU is running in debug mode.
	$conf{"scp"} =~ s/$conf{'scp'}/$conf{'scp'} -v/;
    } 
    else {					# Run scp in quiet mode unless ICU is running in debug or init mode. 
	$conf{"scp"} =~ s/$conf{'scp'}/$conf{'scp'} -q/ unless (defined($opt_i));
    }

  # Replace %stuff% with $conf{'stuff'} in the get_conf/db/bin commands.
    @tmparr = (\$conf{'get_bin_cmd'}, \$conf{'get_conf_cmd'}, \$conf{'get_db_cmd'});
    foreach $_ (@tmparr) {
	$$_ =~ s/%scp%/$conf{'scp'}/g; 
	$$_ =~ s/%port%/$icuhost{'sshport'}/g;
	$$_ =~ s/%tmp_dir%/$tmp_dir/g;
	$$_ =~ s/%key_get_bin_priv%/$conf{'key_get_bin_priv'}/g; 
	$$_ =~ s/%key_get_conf_priv%/$conf{'key_get_conf_priv'}/g;
	$$_ =~ s/%key_get_db_priv%/$conf{'key_get_db_priv'}/g;
	$$_ =~ s/%hostname%/$icuhost{'name'}/g; 
    }

    $debug && print STDERR "Setting timeout to $conf{'getfiles_timeout'} seconds.\n"; 
    alarm($conf{'getfiles_timeout'});  
    $timeout_getfiles = 1;    

    $verbose && print "Getting files from $icuhost{'name'}:";
    $scp_output = undef;

    if ($get_bin) {
	unlink("$conf{tmp_dir}/aide.bin") or icu_die("Could not unlink $conf{tmp_dir}/aide.bin: $!")
	  if (-e "$conf{tmp_dir}/aide.bin");
	getfile("$conf{'get_bin_cmd'}", "aide.bin");
	push(@files,"$tmp_dir/aide.bin");
    }
    if ($get_conf) {
        unlink("$conf{tmp_dir}/aide.conf") or icu_die("Could not unlink $conf{tmp_dir}/aide.conf: $!")
          if (-e "$conf{tmp_dir}/aide.conf");    
	getfile("$conf{'get_conf_cmd'}","aide.conf");
	push(@files,"$tmp_dir/aide.conf");
    }         
    if ($get_db) {
        unlink("$conf{tmp_dir}/aide.db") or icu_die("Could not unlink $conf{tmp_dir}/aide.db: $!")
          if (-e "$conf{tmp_dir}/aide.db");    
	getfile("$conf{'get_db_cmd'}",  "aide.db");
	push(@files,"$tmp_dir/aide.db");	
    }         

    print ".\n" unless (defined($opt_i) || !$verbose);
    alarm(0);
    $timeout_getfiles = 0;

# If scp gave any output - send it in the mail and also write it to the host's log.
    if (defined($scp_output)){  
	notify(1,0,1,1,"Warning: scp gave output when copying files.");
        notify(0,0,0,1,"Here is the output scp gave:\n$scp_output");
        notify(0,0,0,1,"End of scp output.\n");
	get_mail("scp_output") if (defined($scp_output));
	$mail_admin = 1 if ($conf{'mail_to_icu_admin_scp_output'});   
    }

  # Check if the required files was actually received.
    @missing_files = "";
    for ($i = 0; $i <= $#files; $i++) {
	if (-e "$files[$i]") {         
	    $debug && print STDERR "$files[$i] was successfully received.\n";
	} 
	else { 
	    notify(1,0,1,1,"$files[$i] was NOT received.");
	    $files[$i] =~ s/.+\///;  
	    push(@missing_files,"$files[$i] ($conf{'host_basedir'}/$files[$i] on $icuhost{'name'})\n");
	}
    }

    if ($#missing_files > 0) {	# At least one file was not received.
	notify(1,0,1,1,"Critical file(s) missing, aborting.");
	get_mail("files_missing");
	
        if (defined($opt_c)) {  # If running -c, tell if this was before or after the check.
            $mailbody .= "Note: This was after the actual check.\n"      if ($check_done);
            $mailbody .= "Note: This was before the actual check.\n" unless ($check_done);
        }       

	$mailbody .= "Here is the list of missing files:\n@missing_files\n";
	$mail_admin = 1 if ($conf{'mail_to_icu_admin_missing_files'});
	clean_exit;
    } 
    else {
 	$verbose && print "All files successfully received.\n";
    }

    $debug && print STDERR "* sub getfiles ended\n";
}



sub lockfile    # Create or remove lock file.  Syntax: lockfile("create") or lockfile("remove").
{
    $debug && print STDERR "* sub lockfile started\n"; 

    my ($lockfile, $mailto);	

    my $action  = shift;	  				# Action is "create" or "remove".
    $lockfile   = "$conf{'lockfile_dir'}/$icuhost{'name'}.lock";
    $mailto     = $conf{'admin_email'};				# ICU admin always gets this mail. 
    $mailto    .= ",$icuhost{'mailto'}" if ($conf{'lockfile_also_notify_host_admin'});	
    $mail_admin = 0;   # admin_email is already in $mailto now (set so we don't get an empty $mailto).

  # Create new lock file.
    if ($action eq "create") {
	$debug && print STDERR "Checking lock file.\n";

	if (-e "$lockfile") {      				# Exit if a lock file already exists.
	    notify(1,0,1,1,"Job aborted: Lock file found for $icuhost{'name'}: $lockfile\n");
	    $verbose && print "Check/init already in progress?\n";
	    $verbose && print "Investigate the cause and remove the lock file manually.\n";
	    get_mail("lockfile_exists") unless (defined($opt_i)); 
	    if (defined($mailbody)) {
	        if (defined($opt_N)) {
             	    notify(2,0,1,1,"ICU started with the -N switch - printing mail to standard output instead of sending it.");
                    notify(2,0,0,1,"Message subject: $subject");
                    notify(2,0,0,1,"Message body:\n$mailbody");
                    notify(2,0,0,1,"End of message body.");     
		} 
		else {
		    send_mail($mailto,$mail_admin,$subject,$mailbody);
	        }                                   
	    }
	    exit;
        } 
	else {							   # Now really create the new lock file.
	    $debug && print STDERR "Not found. Creating new.\n";
	    open(LOCKFILE,">$lockfile") || icu_die("Could not create lockfile $lockfile: $!\n");
	    print LOCKFILE "This is a lock file for the host $icuhost{'name'}.\n";
	    print LOCKFILE "Process started $start_date, PID: $$.\n";
	    print LOCKFILE "Command line: ICU @args\n";
	    close(LOCKFILE);
	}
    } 
  # Remove lock file.
    elsif ($action eq "remove") {
        $debug && print STDERR "Removing lock file ($lockfile)\n";
   	unlink("$lockfile") or icu_die("Could not remove $lockfile: $!\n")
	  if (-e "$lockfile");
    }

    $debug && print STDERR "* sub lockfile ended\n";
}



sub check_root      # Check if ICU is started as root. 
{
    $debug && print STDERR "* sub check_root started\n";

    if ($conf{'abort_if_root'} && ! $>) { 
        print "You have started $0 as user root but abort_if_root is set to 1 in $icu_config.\n";
	print "Running ICU as root is never recommended.\n";
	print "\nExiting...\n";
	exit;
    }

    $debug && print STDERR "* sub check_root ended\n";
}



sub check_port     # Do a TCP connect to check if given port is open on a host. Syntax: check_port(hostname,port).
{
    $debug && print STDERR "* sub check_port started\n"; 

    my ($hostname, $port, $iaddr, $paddr, $proto);

    $hostname = shift;
    $port     = shift;
    $iaddr    = inet_aton($hostname) or icu_die("Could not resolve hostname $hostname");
    $paddr    = sockaddr_in($port,$iaddr);
    $proto    = getprotobyname('TCP');   

    socket(SOCK,PF_INET,SOCK_STREAM,$proto);
    $verbose && print "Checking port $port...";

    $debug && print STDERR "Setting timeout to $conf{'check_port_timeout'} seconds.\n"; 
    alarm($conf{'check_port_timeout'});
    $timeout_connect = 1;                   # So the beepbeepbeepbeep sub knows what we were trying to do

    if (connect(SOCK,$paddr)) {             # The port is open.
	alarm(0);
	$timeout_connect = 0;
	$verbose && print "open.\n";
	close(SOCK);
	$debug && print STDERR "* sub check_port returning 1\n";
	return 1;
    } 
    else {				    # The port is closed.
	alarm(0);
	$timeout_connect = 0;
	$verbose && print "closed.\n";
	$timeout_connect = 0;
	$mailbody        = undef;
	$subject         = undef;
	$timeout_reason  = undef;
	$errormsg = $!;
	close(SOCK);
	$debug && print STDERR "* sub check_port returning 0\n";
	return 0;
    }

    $debug && print STDERR "* sub check_port ended\n";
}



sub send_mail   #  Syntax: send_mail(to,0/1,subject,body).
		#  0 = Don't send copy to ICU admin, 1 = Also send copy to ICU admin. 
{
    $debug && print STDERR "* sub send_mail started\n"; 

    my (@body, $mailto, $to_icu_admin, $subject, $tmpstr, $date);	

    $mailto       = shift;
    $to_icu_admin = shift; 
    $subject      = shift;
    @body         = @_;
    $mailto      .= ",$conf{'admin_email'}" if ($to_icu_admin);	# Also send to ICU admin?
	
  # Send empty report?
    if ($check_done && $#aide_output == 0 && !$conf{'allow_empty_reports'}) {
 	push(@body,"\nThe filesystem check was executed as normal, and AIDE gave no output.\n");
    }

    $subject = "[ICU - $hostname]  $subject";
    notify(1,0,1,1,"Sending mail to $mailto with subject: $subject");
    $date = (scalar localtime);
    $tmpstr = "$conf{'log_dir'}/$icuhost{'name'}.log";

    unless (defined($opt_i)) {			# Save the mail in the host's log (but not the welcome mail).
	notify(0,0,0,1,"Message body:\n@body");
	notify(0,0,0,1,"End of message body.");
    }

    open(SENDMAIL, "|$conf{'sendmail'} $mailto") or die("SENDMAIL: Could not open sendmail process: $!\n\nExiting");
    print SENDMAIL "From: $conf{'mail_from'} <$conf{'admin_email'}>\n";
    print SENDMAIL "To: $mailto\n";
    print SENDMAIL "Subject: $subject\n";
    print SENDMAIL "X-ICU-version: $version\n";
    print SENDMAIL "@body\n\n";
    close(SENDMAIL);

    $debug && print STDERR "* sub send_mail ended\n";
}


sub get_mail      # Syntax: get_mail("mail"). Result will be appended to $mailbody and $subject.
{
    $debug && print STDERR "* sub get_mail started\n";
    
    my $mailtype = shift;
    $debug && print STDERR "Getting mail ->$mailtype<-\n";

    if ($mailtype eq "cant_connect") {
	$subject  .= "Warning: Could not connect to port $icuhost{'sshport'}";      
	$mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The check/init of $icuhost{'name'} started $start_date was aborted\n";
        $mailbody .= "because the ICU server ($conf{'icu_server_name'}) was unable to connect to\n";
	$mailbody .= "port $icuhost{'sshport'} on $icuhost{'name'}. This could mean that $icuhost{'name'}\n";
        $mailbody .= "isn't running sshd on port $icuhost{'sshport'} or that the host is down.\n";
	$mailbody .= "You should investigate this as soon as possible!\n";
    }

    if ($mailtype eq "filesystem_changed") { 
	$subject  .= "Warning: Filesystem has changed"; 
	$mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The AIDE command gave some output while checking $icuhost{'name'}!\n";
	$mailbody .= "This probably means that the filesystem has changed, which could mean that\n";  
	$mailbody .= "authorized changes were made. It could also mean that the host has been\n";
	$mailbody .= "compromised. The database has been updated with these changes and will now\n";
	$mailbody .= "be regarded as safe.\n"; 
	$mailbody .= "Consider updating your $conf{'host_basedir'}/aide.conf if you get warnings\n";
	$mailbody .= "about these changes all the time but think that the changes are OK.\n";
    }

    if ($mailtype eq "aide_binary_md5_mismatch") {
	$subject  .= "Warning: The AIDE BINARY has changed!! ";
	$mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The AIDE binary ($conf{'host_basedir'}/aide.bin) on $icuhost{'name'}\n";
        $mailbody .= "differs from the AIDE binary stored locally on the ICU server.\n";
	$mailbody .= "THIS SHOULD NEVER HAPPEN and could mean that the host has been\n"; 
	$mailbody .= "compromised. Investigate this as soon as possible!\n";
    }
	
    if ($mailtype eq "aide_db_md5_mismatch") {
        $subject  .= "Warning: The AIDE database file has changed. ";
        $mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The AIDE database file on $icuhost{'name'} differs from\n"; 
	$mailbody .= "the database file stored locally on the ICU server.\n";
        $mailbody .= "This could mean that authorized changes were made,\n";
        $mailbody .= "but it could also mean that the host has been compromised.\n";
	$mailbody .= "The AIDE database on the ICU server has been updated with these changes\n";
	$mailbody .= "and will now be regarded as safe. If only the timestamp has changed\n";
	$mailbody .= "it might mean that a check was run locally or that an earlier ICU check\n";
        $mailbody .= "was aborted for some reason.\n";
    }

    if ($mailtype eq "aide_conf_md5_mismatch") {    
	$subject  .= "Warning: The AIDE configuration file has changed. ";  
	$mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The AIDE configuration file on $icuhost{'name'} differs from\n";
	$mailbody .= "the configuration file stored locally on the ICU server.\n";
	$mailbody .= "This could mean that authorized changes were made,\n";
	$mailbody .= "but it could also mean that the host has been compromised.\n";
	$mailbody .= "The AIDE configuration file on the ICU server has been updated with these\n"; 
	$mailbody .= "changes and will now be regarded as safe.\n"; 
    }

    if ($mailtype eq "files_missing") {
	$subject  .= "Warning: Critical file(s) missing. ";
        $mailbody .= "\n*** Warning ***\n";
        $mailbody .= "At least one critical file was missing on $icuhost{'name'}.\n";
	$mailbody .= "Either the file transfer failed or the files were missing on the remote host.\n";
    }                                                                    

    if ($mailtype eq "getfiles_timeout") {
	$subject  .= "Warning: Could not get files ";
        $mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The check of $icuhost{'name'} started $start_date was aborted\n";
        $mailbody .= "because the file transfer timed out.\n";
        $mailbody .= "You should investigate this as soon as possible!\n";
    }                            

    if ($mailtype eq "check_timeout") {
	$subject  .= "Warning: The AIDE filesystem check timed out. ";
        $mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The check of $icuhost{'name'} started $start_date timed out.\n";
	$mailbody .= "You should investigate this as soon as possible!\n";
    }                         

    if ($mailtype eq "lockfile_exists") {
        $subject  .= "Warning: Lock file found while trying to initialize the filesystem check. ";
        $mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The check of $icuhost{'name'} started $start_date was aborted because a\n";
        $mailbody .= "lock file for this host was found. A check may already be in progress or a\n";
	$mailbody .= "fatal error occurred during an earlier check. The ICU administrator (who has\n";
	$mailbody .= "now been informed) may have to remove the file $conf{'lockfile_dir'}/$icuhost{'name'}.lock\n";
	$mailbody .= "manually from the ICU server if the problem persists.\n";
    }

    if ($mailtype eq "icu_die") {
        $subject  .= "Warning: ICU had to be killed during process. ";
        $mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The check/init of $icuhost{'name'} started $start_date had to be killed.\n";
	$mailbody .= "The ICU administrator has also received this message and should investigate\n";
	$mailbody .= "this as soon as possible!\n";
	$mailbody .= "Possible existing lock file will remain until removed manually.\n";
    }

    if ($mailtype eq "empty_report") {
        $subject  .= "AIDE found no change in the filesystem. ";
        $mailbody .= "The AIDE filesystem check of $hostname started\n";
        $mailbody .= "$start_date found no change in the filesystem.\n";
    }

    if ($mailtype eq "scp_output") {
	$subject  .= " Warning: scp gave output. " unless (defined($subject) && $subject =~ /Warning: scp gave output/);
        $mailbody .= "\n*** Warning ***\n";
	$mailbody .= "The scp command gave output while copying files from $icuhost{'name'}.\n";
        $mailbody .= "It may not be anything critical but you should investigate the cause.\n";
	$mailbody .= "Here is the output scp gave:\n$scp_output\n";
    }

    if ($mailtype eq "invalid_timestamp") {
        $subject  .= "Invalid database creation timestamp after check. ";
        $mailbody .= "\n*** Warning ***\n";
	$mailbody .= "Whenever AIDE creates the new database after a check, it puts a timestamp\n";
	$mailbody .= "in it. ICU grabs the old timestamp before the check and compares it to\n";
	$mailbody .= "the new timestamp after the check. The new date/time should obviously\n";
	$mailbody .= "always be more recent than it was before the check, but this time it was NOT!\n"; 
        $mailbody .= "If the timestamps are exactly the same it may mean that the AIDE command\n";
	$mailbody .= "failed for some reason. If the timestamp after the check is OLDER than the\n";
	$mailbody .= "timestamp before the check it may mean that someone is doing something\n";
        $mailbody .= "really nasty to your host. You should investigate this as soon as possible!\n";   
    }

    if ($mailtype eq "welcome") {
	$subject = $conf{'welcome_mail_subject'};
	open(WELCOME,$conf{'welcome_mail_body'}) 
          or die("WELCOME: Could not open $conf{'welcome_mail_body'}: $!\n\nExiting");
	while(<WELCOME>) {
            $_ =~ s/%hostname%/$icuhost{'name'}/g;
	    $_ =~ s/%admin_email%/$conf{'admin_email'}/g;
	    $_ =~ s/%host_basedir%/$conf{'host_basedir'}/g;
	    $mailbody .= $_;
	}
	close(WELCOME);
    }

    $debug && print STDERR "* sub get_mail ended\n";
}



sub asciifilecmp      # Compare two ascii files and return differences. 
                      # Syntax: ($added,$removed) = asciifilecmp(oldfile,newfile).
{
    $debug && print STDERR "* sub asciifilecmp started\n";  
 
    my ($oldfile, $newfile, @oldarr, @newarr, $added, $removed, $i, $j);

    $oldfile = shift;
    $newfile = shift;
 
    open(OLDFILE,"$oldfile") or icu_die ("OLDFILE: could not open $oldfile: $!");
    @oldarr = <OLDFILE>;
    close(OLDFILE);

    open(NEWFILE,"$newfile") or icu_die ("NEWFILE: could not open $newfile: $!");
    @newarr = <NEWFILE>;
    close(NEWFILE);     
                      
    OUT:for ($i = 0; $i <= $#oldarr; $i++) {
	IN:for ($j = 0; $j <= $#newarr; $j++) {
            next IN unless (defined($newarr[$j]));
            if ($oldarr[$i] eq $newarr[$j]) {
                $oldarr[$i] = undef;
                $newarr[$j] = undef;
                next OUT;
            }
        }
    }

    foreach $_ (@oldarr) {	     # Remaining lines in @oldarr have been removed.
	if (defined($_)) {
	    chomp($_);
            $removed .= "$_\n";
	}
    }

    foreach $_ (@newarr) {	     # Remaining lines in @newarr have been added.
	if (defined($_)) {
	    chomp($_);
            $added .= "$_\n";
	}
    }

    $debug && print STDERR "* sub asciifilecmp returning";
    return($added,$removed);
}                            



sub tmp_dir   # Create or remove temporary directory for a host. Syntax: tmp_dir("new") or tmp_dir("remove").
{
    $debug && print STDERR "* sub tmp_dir started\n"; 

    my ($basetmp, $error);
    my $action = shift;

  # Current $tmp_dir should always be deleted when sub tmp_dir is called.
    if (-e "$tmp_dir") {
	$debug && print STDERR "Removing temporary directory ($tmp_dir).\n";
	unlink(<$tmp_dir/*>);
        if (-e "$icuhost{'name'}.icu-install") {
            unlink(<$tmp_dir/$icuhost{'name'}.icu-install/*>);
            rmdir("$icuhost{'name'}.icu-install") or icu_die("Could not remove $tmp_dir: $!");
        }
        rmdir("$tmp_dir") or icu_die("Could not remove $tmp_dir: $!");
    }      

    if ($action eq "new") {  	 	# Create new temporary directory.
	$debug && print STDERR "Creating new temporary directory ($tmp_dir).\n";
	mkdir("$tmp_dir",0700) or icu_die("Could not create directory $tmp_dir: $!");
    }

    $debug && print STDERR "* sub tmp_dir ended\n";
}



sub clean_exit    # Send/print mail, write to logs, remove temporary files, remove lock file and exit.
{
    $debug && print STDERR "* sub clean_exit started\n";

    my ($tmpstr, $date);

    $debug && print STDERR "* sub clean_exit started\n";
    if (defined($mailbody)) {				# There is a mail to send.
        unless ($check_done || defined($opt_i)) {       # Give an extra note if no check has been done.
            $_ = "[THE CHECK WAS ABORTED!!]";
            $subject  .= " $_";
            $mailbody .= "\n$_\n";
        }           
	if (defined($opt_N) || defined($opt_i)) {	# Print mail to stdout if -n or -i.
	    if (defined($opt_i)) {			# If $mailbody is defined here and we're running -i it
		notify(2,0,0,1,"$mailbody");            # means that a fatal error occured (like missing files).
		notify(2,1,1,1,"Unexpected output during init => database initialization FAILED!");
	    }
	    else {
		notify(2,0,1,1,"ICU started with the -N switch - printing mail to standard output instead of sending it.");
		notify(2,0,0,1,"Message subject: $subject");
		notify(2,0,0,1,"Message body:\n$mailbody");
		notify(2,0,0,1,"End of message body.");
	    }
	} 
	else {
 	    send_mail($icuhost{'mailto'},$mail_admin,$subject,$mailbody);    # Send the mail.
	}
    }

  # When -i, Send the welcome mail and only the welcome mail to the host admin. 
  # (other possible output was printed to stdout above)
    if (!defined($mailbody) && defined($opt_i) && $conf{'send_welcome_mail'}) {
	if (defined($opt_N)) {
	    notify(2,0,1,1,"ICU started with the -N switch - not sending any welcome mail.\n");
	}
	else {
	    get_mail("welcome");
	    send_mail($icuhost{'mailto'},$mail_admin,$subject,$mailbody);
	}
    }

    notify(1,0,1,1,"Check ended.")                   if (defined($opt_c));
    notify(1,0,1,1,"Database initialization ended.") if (defined($opt_i));
    tmp_dir("remove");
    lockfile("remove");	
    $debug && print STDERR "* sub clean_exit ended\n";
    select(STDOUT);
    exit;
}



sub notify 	# Write string to logs and also automatically to stderr if running in verbose mode. 
                # Syntax: notify(x,y,z,"string").
                # x=printopt, y=log_icu, z=log_host
{
    $debug && print STDERR "* sub notify started\n"; 

    my ($date, $tmpstr, @tmparr, $msgline, @msglines, $hostname);   

    my $printopt = shift;  # Print to stdout/stderr? 0=never, 1=only in verbose mode, 2=always
    my $output   = shift;  # If $printopt == 1, print to stdout or stderr? (0=stdout, 1=stderr)
    my $log_icu  = shift;  # should we log to ICU.log/syslog?
    my $log_host = shift;  # Should we log to the host log?
    my $message	 = shift;  # The actual message.

    $date = (scalar localtime);
    chomp($message);
    @msglines = split(/\n/, $message);  	# Split the message so we can add a prefix to each line.

    if (defined($icuhost{'name'})) {
	$hostname = $icuhost{'name'};
    }
    else {
	$hostname = "";
    }

  # Print to stdout/stderr if both $verbose and $printopt are set, or if $printopt is set to 2.
    if (($verbose && $printopt) || $printopt == 2) {
	print STDOUT "$message\n" if ($output == 0);	# Print to stdout
	print STDERR "$message\n" if ($output == 1);	# Print to stderr
    }

  # Write to ICU.log/syslog if $icu_log is set.
    if ($log_icu) {
	$tmpstr = "$conf{'log_dir'}/ICU.log";
	open(ICULOG,">>$tmpstr") or die("ICULOG: Could not open $tmpstr for writing: $!\n\nExiting");
	flock(ICULOG, LOCK_EX) or print STDERR "Warning, unable to lock ICU.log: $!\n";

        foreach $msgline(@msglines) {
            print ICULOG "$date ICU[$$] $hostname: $msgline\n";
        }

	flock(ICULOG, LOCK_UN) or print "Warning, unable to unlock ICU.log: $!\n"; 
	close(ICULOG);

      # Same thing goes to syslog if $use_syslog is set to 1.
	foreach $msgline(@msglines) {
	    system("$conf{'logger'} $conf{'logger_arguments'} -t \"ICU[$$] $hostname\" \"$msgline\n\" >/dev/null 2>&1") 
              if ($conf{'use_syslog'});
	}
    }

  # Write to the host log if $log_host is set.
    if ($log_host && defined($icuhost{'name'})) {
	$tmpstr = "$conf{'log_dir'}/$icuhost{'name'}.log";
	open(HOSTLOG,">>$tmpstr") or die("HOSTLOG: Could not open $tmpstr for writing: $!\n\nExiting");

        foreach $msgline(@msglines) {
            print HOSTLOG "$date ICU[$$] $icuhost{'name'}: $msgline\n";
        }

        close(HOSTLOG);
    }

    $debug && print STDERR "* sub notify ended\n";
}



sub beepbeepbeepbeep    	    # Take care of the alarm (something timed out).
{
    $debug && print STDERR "* sub beepbeepbeepbeep started\n";

    get_mail("scp_output") if (defined($scp_output));

    if ($timeout_connect) {	    # Timeout during port check.
	$verbose && print "timeout.\n";
	get_mail("cant_connect");
	$mailbody .= "Additional information: connection timed out after $conf{'check_port_timeout'} seconds.\n";
	$subject  .= "- connection timed out after $conf{'check_port_timeout'} seconds.";
	$timeout_reason = "Could not connect to port $icuhost{'sshport'}: connection timed out after $conf{'check_port_timeout'} seconds.\n";
	$mail_admin = 1 if ($conf{'mail_to_icu_admin_port_closed'});
    }
    elsif ($timeout_checking) {			
	get_mail("check_timeout");  # Timeout during filesystem check.
	$mailbody .= "Additional information: Check timed out after $conf{'check_timeout'} seconds.\n";
	$timeout_reason = "Check timed out after $conf{'check_timeout'} seconds.\n";                        
	$mail_admin = 1 if ($conf{'mail_to_icu_admin_check_timeout'});
    }
    elsif ($timeout_getfiles) {	    # Timeout during file tranfer.
	$verbose && print " - timeout.\n";
	get_mail("getfiles_timeout");
	$mailbody .= "Additional information: File transfer timed out after $conf{'getfiles_timeout'} seconds.\n";

        if (defined($opt_c)) {      # If running -c, tell if this was before or after the check.
            $mailbody .= "Note: This was after the actual check.\n"      if ($check_done);
            $mailbody .= "Note: This was before the actual check.\n" unless ($check_done);
        }     

	$subject  .= "- file transfer timed out after $conf{'getfiles_timeout'} seconds.";
	$timeout_reason = "File transfer timed out after $conf{'getfiles_timeout'} seconds.\n";       
	$mail_admin = 1 if ($conf{'mail_to_icu_admin_getfiles_timeout'});
    }

    notify(1,0,1,1,$timeout_reason);
    clean_exit;             
}



sub icu_die     # Write to logs, send mail and exit. Don't remove the lock file because a weird
		# error that needs to be investigated has probably occurred when this sub is called.
{
    $debug && print STDERR "* sub icu_die started\n";

    my $errmsg = shift;
    chomp($errmsg);

    notify(1,1,1,1,"Fatal error: $errmsg, Exiting...");

    if (defined($opt_c)) {
	get_mail("icu_die");
	$mail_admin = 1;
        $mailbody .= "Additional information: $errmsg\n";
        $mailbody .= "Note: This was after the actual check.\n"      if ($check_done);
        $mailbody .= "Note: This was before the actual check.\n" unless ($check_done);
    }
    else {
	$opt_N = 1;	# Don't send mail unless running -c.
    }

    if (defined($opt_N)) {
	print $mailbody if(defined($mailbody));
    }
    else {
	send_mail($icuhost{'mailto'},$mail_admin,$subject,$mailbody);
    }         

    # Exit without removing lock file/temporary files because you migh
    # want to take a look at them.
    exit;
}



sub sanity_check     # Check for required files and their permissions etc.
{
    my 
       (
         $output, $tmpstr, @tmparr, @dirfiles, $file, $perm, $s_verbose, 
         $date, $mailto, $required, $failbanner, $mode, $real_file
       );

    $debug && print STDERR "* sub sanity_check started\n"; 

    if (($verbose && defined($opt_s)) || $debug) {
	$s_verbose = 1;			# Be extra verbose when running -s -v
    }
    else {
	$s_verbose = 0;
    }

    $verbose && print "Initializing sanity check.\n";

    $failbanner = "Sanity check failed for command '$0 @args' executed $start_date.\n";

  # Parameters that must be set in ICU config. Exit if any parameter is missing.
    $s_verbose && print "- Checking if all options are defined in $icu_config.\n";
    my @conf_req = qw 
      ( 
	icu_basedir icu_server_name admin_email md5sum ssh scp ssh_keygen tar logger sendmail
	cp compress use_syslog send_welcome_mail mail_to_icu_admin_db_md5 mail_to_icu_admin_conf_md5
	mail_to_icu_admin_check_timeout allow_empty_reports compress_databases lockfile_also_notify_host_admin
	mail_to_icu_admin_port_closed continue_if_binary_changed abort_if_root getfiles_timeout
        check_timeout check_port_timeout host_basedir mail_from icuhosts log_dir bin_dir config_dir
        database_dir icu_client_installscript welcome_mail_body readme_installscript ignore_traverse_tree_warnings
        welcome_mail_subject key_init_db_pub key_init_db_priv key_check_pub key_check_priv silent_suppressing
        key_get_conf_pub key_get_conf_priv key_get_bin_pub key_get_bin_priv key_get_db_pub
        key_get_db_priv aide_check_cmd aide_init_db_cmd check_cmd init_db_cmd send_bin_cmd
        send_conf_cmd send_db_cmd get_bin_cmd get_conf_cmd get_db_cmd mail_to_icu_admin_scp_output
        mail_to_icu_admin_missing_files ignore_future_atimes ignore_future_ctimes ignore_future_mtimes
	mail_icu_admin_invalid_timestamp mail_to_icu_admin_getfiles_timeout
      );	

    foreach $required (@conf_req) {
	unless (defined($conf{"$required"})) {
	    $output .= "* The parameter '$required' is not defined anywhere in $icu_config.\n";
	}	
    }

    if (defined($output)) {    # Something failed - don't even continue the sanity checks.
	print STDERR "$failbanner$output\nExiting...\n";
	exit;
    }
    else {
	$debug && print STDERR "Found all required options in $icu_config.\n";
    }

  # Check if we're running as root and if we should quit.
    check_root;

  # Binaries that must exist and be executable, but don't bother if they aren't readable 
  # by the ICU user. Checking if they are readable may mess up some things, for example
  # the ssh binary is usually suid root and has these permissions: -rws--x--x.
    @tmparr = 
      (
	 $conf{'scp'}, $conf{'md5sum'}, $conf{'ssh'}, $conf{'tar'}, $conf{'sendmail'},
	 $conf{'ssh_keygen'}, $conf{'cp'}
      );
    foreach $tmpstr (@tmparr) {
	unless (-e "$tmpstr" && -x "$tmpstr") {  
	    $output .= "* $tmpstr doesn't exist.\n";
	}
    }

  # Files that must exist and be readable.
    $s_verbose && print "- Checking files that should exist and be readable by you.\n";
    @tmparr = 
      (
	 $conf{'database_dir'}, $conf{'key_init_db_pub'}, $conf{'key_init_db_priv'}, $conf{'key_check_pub'},
	 $conf{'icu_basedir'}, $conf{'tmp_dir'}, $conf{'icuhosts'}, $conf{'bin_dir'}, $conf{'config_dir'},  
 	 $conf{'log_dir'}, $conf{'key_check_priv'}, $conf{'key_get_conf_pub'}, $conf{'key_get_conf_priv'},
         $conf{'key_get_bin_pub'}, $conf{'key_get_bin_priv'}, $conf{'key_get_db_pub'},$conf{'readme_installscript'}, 
  	 $conf{'key_get_db_priv'}, $conf{'icu_client_installscript'},
      );
    foreach $tmpstr (@tmparr) {
        unless (-e "$tmpstr" && -r "$tmpstr") {
            $output .= "* $tmpstr doesn't exist or isn't readable by you.\n";
        }
    }            

  # Make sure all aide.conf* files in config_dir are readable.
    $s_verbose && print "- Checking if all AIDE configurations in $conf{'config_dir'} are readable.\n";
    if (-e "$conf{'config_dir'}") {		
	opendir(DIR,"$conf{'config_dir'}");
        @dirfiles = readdir(DIR);
	foreach $file (@dirfiles) {
	    if ($file =~ /^aide\.conf/ && !-r "$conf{'config_dir'}/$file") { 
		$output .= "* $conf{'config_dir'}/$file exists but is not readable by you.\n";  
	    }
        }
    }

  # Make sure all aide.bin* files in bin_dir are readable.
    $s_verbose && print "- Checking if all AIDE binaries in $conf{'bin_dir'} are readable.\n";
    if (-e "$conf{'bin_dir'}") {		
        opendir(DIR,"$conf{'bin_dir'}");
        @dirfiles = readdir(DIR);
        foreach $file (@dirfiles) {
            if ($file =~ /^aide\.bin/ && !-r "$conf{'bin_dir'}/$file") {
                $output .= "* $conf{'bin_dir'}/$file exists but is not readable by you.\n";
            }
        }
    }                  

  # Files that must be owned and writable by user.
    $s_verbose && print "- Checking files that should be owned and writable by you.\n";
    @tmparr = ($conf{'icu_basedir'}, $conf{'log_dir'}, $conf{'tmp_dir'}, $conf{'database_dir'});
    OWN_WRITE:foreach $tmpstr (@tmparr) { 
	next unless (-e "$tmpstr");   # Already checked this above
	unless (-o $tmpstr) {         # Not owned by user?
	    $output .= "* $tmpstr is not owned by you.\n";
	}
	unless (-w $tmpstr) {         # Not writable by user?
	    $output .= "* $tmpstr is not writable. Use 'chmod u+w $tmpstr' to fix it.\n"; 
	    next OWN_WRITE;
	}
	if (-d $tmpstr && !-x $tmpstr) {
	    $output .= "* $tmpstr should be a writable directory and must be set +x. Use 'chmod u+x $tmpstr' to fix it.\n";
	}
    }

  # Make sure the permissions on these files or directories are ok.
  # Owner must have rwx access, and group may have it.
    $s_verbose && print "- Checking permissions of specific files/directories.\n";
    @tmparr = ($conf{'icu_basedir'}, $conf{'tmp_dir'}, $conf{'log_dir'}, $conf{'database_dir'}, $conf{'config_dir'});

    PERMCHECK: foreach $tmpstr (@tmparr) {
	next unless (-e "$tmpstr");
	$mode = (stat("$tmpstr"))[2];
	$perm = sprintf "%04o", $mode & 07777;
	unless ($perm eq "0700") {

            if (-l "$tmpstr") {   		   # It's a symlink so print what it points to.
	        $real_file = readlink("$tmpstr");
	    }
	    else {
	        $real_file = $tmpstr;	
	    }

	    $output .= "* $real_file has incorrect permissions.\n";
	    $output .= "  It's currently $perm but should be 0700. Try 'chmod 0700 $real_file' to fix it.\n";
        }
    }

    $s_verbose && print "- Checking if additional binaries required by non-mandatory options exist.\n";

  # If use_syslog is set to 1, make sure the logger binary exists.
    if ($conf{'use_syslog'} && !-e "$conf{'logger'}") {
	$output .= "* You have set use_syslog=1 in $icu_config but the logger binary $conf{'welcome_mail_body'} was not found.\n";
    }

  # If compress_databases is set to 1, make sure the compress program exists.
    if ($conf{'compress_databases'} && !-e "$conf{'compress'}") {
	$output .= "* You have set compress_databases=1 in $icu_config but the compress binary ($conf{'compress'}) was not found.\n";
    }

  # If $conf{'send_welcome_mail'} is set to 1, make sure the mail exists.
    if ($conf{'send_welcome_mail'} &&  !-e "$conf{'welcome_mail_body'}") {
	$output .= "* You have set send_welcome_mail=1 in $icu_config but the mail $conf{'welcome_mail_body'} was not found.\n"; 
    }

  # Do some tests on the host if user specified -r <hostname>.
    if (defined($opt_r)) {
	$s_verbose && print "- Trying to find $hostname in $conf{'icuhosts'}.\n";
	if (get_host_entry("$hostname")) {   # Does it exist in the host config?
	    print STDERR "  Found: mailto=$icuhost{'mailto'}, OS=$icuhost{'os'}, architecture=$icuhost{'arch'}, port=$icuhost{'sshport'}.\n" 
              if $s_verbose;

	    $s_verbose && print "- Checking if binary/configuration exists for this OS/architecture.\n";
	    unless (-e "$conf{'bin_dir'}/aide.bin-$icuhost{'arch'}-$icuhost{'os'}") {   # No binary for this OS/architecture?
	        $output .= "* No binary found for this OS/architecture (should be found as $conf{'bin_dir'}/aide.bin-$icuhost{'arch'}-$icuhost{'os'}).\n";
            }
	    unless (-e "$conf{'config_dir'}/aide.conf-$icuhost{'arch'}-$icuhost{'os'}") {   # No config found for this OS/architecture?
		$output .= "* No configuration file found for this OS/architecture.  (Should be found as $conf{'config_dir'}/aide.conf-$icuhost{'arch'}-$icuhost{'os'})\n";           
            }
        }
        else {				      # The host was not found in the host config.
	    $output .= "* The host $hostname was not found in $conf{'icuhosts'}.\n";
	}
    }

    if (defined($opt_n)) {   # $conf{'database_dir'}/$hostname must not exist when running -n.
	$tmpstr = "$conf{'database_dir'}/$hostname";
	if (-e "$tmpstr") {
	    $output .= "* The directory $tmpstr already exists. (Tarball already created?)\n";
	    $output .= "  Delete this directory if you really want to create a new tarball.\n";
	} 
    }                      
	
    if (defined($opt_i)) {    # $conf{'database_dir'}/$hostname/archive must exist when running -i.
	$tmpstr = "$conf{'database_dir'}/$hostname/archive";
	unless (-e "$tmpstr") {
	    $output .= "* $tmpstr doesn't exist. (Did you forget to run $0 with the -n switch first?)\n"; 
	}
	$tmpstr = "$conf{'database_dir'}/$hostname/aide.db.current";
	if (-e "$tmpstr") {
	    $output .= "* $tmpstr already exists. (Database already initialized?)\n";
	    $output .= "  Remove this file if you really want to initialize a new database.\n";
	}
	$tmpstr = "$conf{'database_dir'}/$hostname/aide.conf.current";
        if (-e "$tmpstr") {
            $output .= "* $tmpstr already exists. (Database already initialized?)\n";
            $output .= "  Remove this file if you really want to initialize a new database.\n";
        }       
    }
		
    if (defined($opt_c)) {    # $conf{'database_dir'}/$hostname/aide.conf.current must exist when running -c.
        $tmpstr = "$conf{'database_dir'}/$hostname/aide.conf.current";
	unless (-e "$tmpstr") {
	    $output .= "* $tmpstr doesn't exist. (Did you forget to run $0 with the -i switch first?)\n";
	}                        
	$tmpstr = "$conf{'database_dir'}/$hostname/aide.db.current";
	unless (-e "$tmpstr") {    # And aide.db must also exist.
	    $output .= "* $tmpstr doesn't exist. (Did you forget to run $0 with the -i switch first?)\n";
	}
    }	

  # Do a write test in the temporary directory.
    if (-e "$conf{'tmp_dir'}" && -w "$conf{'tmp_dir'}" && -x "$conf{'tmp_dir'}") { 
	$s_verbose && print "- Doing write test in $conf{'tmp_dir'}.\n";
	open(FOO,">$conf{'tmp_dir'}/foo.$$") or $output .= "* Test write in temporary directory failed: could not create $conf{'tmp_dir'}/foo.$$: $!\n";
	close(FOO); 
	unlink("$conf{'tmp_dir'}/foo.$$") if (-e "$conf{'tmp_dir'}/foo.$$");
    }

  # Do a MD5 fingerprint test - make sure we actually get a 32 char output.
    if (-e "$conf{'md5sum'}" && -e "$conf{'cp'}") {
	$s_verbose && print "- Performing MD5 calculation test.\n";
        $tmpstr = `$conf{'md5sum'} $conf{'cp'}`;
        unless ($tmpstr =~ /([\da-f]{32})/) {
 	    $output .= "* The md5sum command ($conf{'md5sum'}) failed - could not find MD5 fingerprint in output. ";
	    $output .= "  The output was: $tmpstr";
	}
    }

    if (defined($output)) {    			     # Something failed - print errors to STDERR and logs.
	$date   = (scalar localtime);
	$tmpstr = "$conf{'log_dir'}/ICU.log";

	$s_verbose && print "\n";
        print STDERR "$failbanner";
        print STDERR "Maybe you need to edit $icu_config? Things that failed:\n$output\nExiting...\n";

	if (-e "$conf{'log_dir'}" && -o "$conf{'log_dir'}" && -w "$conf{'log_dir'}") {
	    if (defined($icuhost{'name'})) {
		notify(0,1,1,1,"$failbanner$output");  # Write to all logs if hostname was defined
	    }
	    else {
		notify(0,1,1,0,"$failbanner$output");  # Skip host log
	    }
	}

	exit;
    } 

    print "Sanity check passed.\n" if (defined($opt_s)) || $verbose;     
    exit if (defined($opt_s));
    $debug && print STDERR "* sub sanity_check ended\n";  
}



sub generate_keys	# Call the sub routine &generate_key with different arguments
			# for each key we need to create.
{
    $debug && print STDERR "* sub generate_keys started\n";  

    my (@files, $file, $key_command);

  # Abort if ICU is started as root?
    check_root;

  # Files/directories that must exist.
    @files = ( $conf{'icu_basedir'}, $conf{'ssh_keygen'}, $conf{'tmp_dir'});
    foreach $file (@files) {
         die("$file doesn't exist.\n\nExiting") unless (-e "$file");
    }                          
 
    print "NOTE: Make sure your key generation binary specified in $icu_config\n";
    print "($conf{'ssh_keygen'}) can generate SSH protocol 1 keys (RSA).\n\n";
    sleep 2;

    unlink(<$conf{'tmp_dir'}/tempkey*>);

  # Generate init_db key.
    $key_command = "command=\"$conf{'aide_init_db_cmd'}\""; 
    generate_key($conf{'key_init_db_priv'},$conf{'key_init_db_pub'},'icu_init_db',$key_command); 

  # Generate check key.
    $key_command = "command=\"$conf{'aide_check_cmd'}\"";
    generate_key($conf{'key_check_priv'},$conf{'key_check_pub'},'icu_check',$key_command); 

  # Generate send_bin key.
    $key_command = "command=\"$conf{'send_bin_cmd'}\"";
    generate_key($conf{'key_get_bin_priv'},$conf{'key_get_bin_pub'},'icu_get_bin',$key_command); 

  # Gernerate send_conf key.
    $key_command = "command=\"$conf{'send_conf_cmd'}\"";
    generate_key($conf{'key_get_conf_priv'},$conf{'key_get_conf_pub'},'icu_get_conf',$key_command); 

  # Gernerate send_db key.
    $key_command = "command=\"$conf{'send_db_cmd'}\"";
    generate_key($conf{'key_get_db_priv'},$conf{'key_get_db_pub'},'icu_get_db',$key_command); 

    unlink(<$conf{'tmp_dir'}/tempkey*>);

    print "\n";
    notify(2,0,1,0,"Key generation successfully completed.\n");
    print "Make sure you are happy with the commands in the public keys (*.pub)\n";
    print "before you distribute them (i.e. add ICU clients).\n";
    print "If you're not happy with the commands, edit $icu_config, remove the\n";
    print "old keys and run $0 -G again. Remember that the keys not ending\n";
    print "in \".pub\" are private, and should be treated as such.\n";
    
    $debug && print STDERR "* sub generate_keys ended\n"; 
    exit;
}



sub generate_key    # Syntax: generate_key(privkey,pubkey,comment,command).
{
    $debug && print STDERR "* sub generate_key started\n"; 

    my ($tmpfile,$tmpfile2,$key,$dir,$fromfile,$tofile,$error);
    my $privkey_filename = shift;
    my $pubkey_filename  = shift;
    my $comment          = shift;
    my $key_command      = shift;

  # Don't overwrite existing keys.
    $error  = "already exists.\nI refuse to overwrite your existing keys.\n";
    $error .= "Make sure your old keys are not used by any client and remove them\n";
    $error .= "if you really want to create new ones.\n";
    die("$privkey_filename $error\nExiting") if (-e $privkey_filename);
    die("$pubkey_filename $error\nExiting")  if (-e $pubkey_filename);  

    print "Generating $privkey_filename and $pubkey_filename.\n";   

  # Remove old temporary keys if they exist.
    $tmpfile = "$conf{'tmp_dir'}/tempkey";
    if (-e "$tmpfile") {
	unlink("$tmpfile") or die("Could not remove temporary file $tmpfile: $!\n\nExiting");
    }
    $tmpfile = "$conf{'tmp_dir'}/tempkey.pub";
    if (-e "$tmpfile") {
        unlink("$tmpfile") or die("Could not remove temporary file $tmpfile: $!\n\nExiting");
    }                  

  # Generate tempkey and tempkey.pub.
    system("$conf{'ssh_keygen'} -b 1024 -C $comment -f $conf{'tmp_dir'}/tempkey -N '' >/dev/null 2>&1");  

  # Make sure tempkey and tempkey.pub now exist.
    $tmpfile  = "$conf{'tmp_dir'}/tempkey";
    $tmpfile2 = "$conf{'tmp_dir'}/tempkey.pub";
    die("Key generation failed: failed to generate RSA key pair $tmpfile and $tmpfile2.\n\nExiting")
      unless (-e "$tmpfile" && -e "$tmpfile2");

  # Open public key tempkey.pub and add the forced command to it.
    $tmpfile = "$conf{'tmp_dir'}/tempkey.pub";
    open(PUBKEY, "$tmpfile") 
      or die "could not open $tmpfile: $!\nExiting";
    $key = <PUBKEY>;
    close(PUBKEY);

    $key = "$key_command $key";				    # Add forced command.

    ($dir) = ($pubkey_filename =~ /(.*\/)/);  	            # Extract pub. key directory.
    die("Unable to extract directory from $pubkey_filename, check your 'key' parameters in $icu_config.\n\nExiting")
      unless (defined($dir));
    ($error = `mkdir -p "$dir" 2>&1`) && die("'mkdir -p $dir': $error\nExiting")
      unless (-e "$dir");				    # Create pub. key directory unless it already exists.

    ($dir) = ($privkey_filename =~ /(.*\/)/);               # Extract priv. key directory.
    die("Unable to extract directory from $pubkey_filename, check your 'key' parameters in $icu_config.\n\nExiting")
      unless (defined($dir));
    ($error = `mkdir -p "$dir" 2>&1`) && die("'mkdir -p $dir': $error\nExiting")
      unless (-e "$dir");				    # Create priv. key directory unless it already exists.

  # Save the new public key.
    open(NEWPUBKEY, ">$pubkey_filename")
      or die "Key generation failed, could not open $pubkey_filename for writing: $!\n\nExiting";
    print NEWPUBKEY $key;
    close(NEWPUBKEY);                             

  # Move the private key to its right directory. 
    $fromfile = "$conf{'tmp_dir'}/tempkey";
    $tofile   = "$privkey_filename";
    rename("$fromfile","$tofile") 
      or die("Fatal error: rename $fromfile to $tofile: $!\n\nExiting"); 

    $debug && print STDERR "* sub generate_key ended\n";
}



sub get_timestamp     # Get creation timestamp from database. Syntax: get_timestamp(dbfile)
{
    $debug && print STDERR "* sub get_timestamp started\n";

    my ($date, $time, $tmpstr);  
    my $dbfile = shift;

    $debug && print STDERR "Getting timestamp from $dbfile.\n";
   
    open(DB,$dbfile) or icu_die("get_timestamp: could not open $dbfile: $!");
    $tmpstr = <DB>;
    $tmpstr = <DB>;  					# Timestamp is on the second line in the database.
    close(DB);

    if (defined($tmpstr) && $tmpstr =~ /^# Time of generation was (20\d{2}-\d{2}-\d{2}) (\d{2}:\d{2}:\d{2})\n$/) {
	$debug && print STDERR "Correct timestamp entry found in $dbfile: $tmpstr";
        $date = $1;
        $time = $2;
        $date =~ s/\D//g;				# Only want the digits.
        $time =~ s/\D//g;
        return($date,$time);
    }
    else {
	$debug && print STDERR "No correct timestamp entry found.\n"; 
    }
 
    $debug && print STDERR "* sub get_timestamp ended\n";
}

# EOF
